import os
import re
from collections import namedtuple
from functools import wraps
from inspect import currentframe
from types import FrameType
from typing import Optional

from .scanner import get_all_blocks
from .scanner import get_variables


class FrameFinder:
    _hierarchy = {
        'self'              : 1,
        'parent'            : 2,
        'grand_parent'      : 3,
        'great_grand_parent': 4,
    }
    
    _frame_0: Optional[FrameType] = None  # the direct caller frame
    _frame_x: Optional[FrameType] = None  # the target caller frame
    
    def getframe(self, func):
        @wraps(func)
        def _wrap(*args, **kwargs):
            frame = self._frame_0 = currentframe().f_back
            for _ in range(self._hierarchy.get(
                    (h := kwargs.get('h', 'self')), h
            ) - 1):
                frame = frame.f_back
            self._frame_x = frame
            return func(*args, **kwargs)
        
        return _wrap
    
    def getinfo(self):
        assert self._frame_0 and self._frame_x, (
            'You must hold the frame before fetching the frame! '
            '(see `FrameFinder.getframe` decorator)'
        )
        struct = namedtuple('FrameInfo', (
            'direct_filename', 'direct_lineno',
            'source_filename', 'source_lineno',
            'source_name',
        ))
        
        def _fix_filename_location(filename: str):
            # warning: in some rare situation that `<frame>.f_code.co_filename`
            # shows a RELATIVE path rather than absolute one. moreover, the
            # 'relative' path is not always corresponded with `os.getcwd`
            # (especially when we have changed working dir by `os.chdir`
            # operations), so we can't make sure `os.path.abspath(os.getcwd()
            # + '/' + <relative_path>)` could always fix it.
            # if fix failed, return '<unknown_root:...>'. the caller should
            # catch this situation and think about how to handle it. see
            # `SourceMap.get_frame_info` and `SourceMap._indexing_filemap`.
            # reference links:
            #   https://bugs.python.org/issue18307
            #   https://github.com/cython/cython/issues/2543
            if filename.startswith('<'):
                # e.g. filename = '<string>', that means the caller frame came
                # from `eval(<string>)` or `exec<string>`.
                return filename
            if os.path.isabs(filename):
                return filename
            elif (x := os.path.join(os.getcwd(), filename)) \
                    and os.path.exists(x):
                return x
            else:
                return f'<unknown_root:{filename}>'
        
        return struct(
            _fix_filename_location(self._frame_0.f_code.co_filename),
            self._frame_0.f_lineno,
            _fix_filename_location(self._frame_x.f_code.co_filename),
            self._frame_x.f_lineno, self._frame_x.f_code.co_name,
        )
    
    @property
    def frame(self):
        return self._frame_0


class SourceMap:
    
    def __init__(self):
        self._sourcemap = {}
        #   dict[filename, dict[lineno, tuple[varname, ...]]]
        self.working_dir = os.getcwd()
        # # self.working_dir = os.path.dirname(os.path.abspath(sys.argv[0]))
    
    def get_frame_info(self, advanced=False):
        info = frame_finder.getinfo()

        # see `FrameFinder.getinfo._fix_filename_location`
        if info.source_filename.startswith('<'):
            filename = info.source_filename
        else:
            filename = os.path.relpath(info.source_filename, self.working_dir)
        lineno = info.source_lineno
        name = x if (x := info.source_name).startswith('<') else x + '()'
        varnames = ()
        
        if advanced:
            if info.direct_filename not in self._sourcemap:
                self._indexing_filemap(info.direct_filename)
            varnames = self._sourcemap[info.direct_filename].get(
                info.direct_lineno, ())
        
        struct = namedtuple('InfoStruct', (
            'filename', 'lineno', 'name', 'varnames'
        ))
        return struct(filename, lineno, name, varnames)
    
    def _indexing_filemap(self, filename: str):
        """
        Args:
            filename
        """
        if filename.startswith('<'):
            # see `FrameFinder.getinfo._fix_filename_location`
            self._sourcemap.setdefault(filename, {})
            return
        
        from .scanner.const import VARIABLE_NAME
        from .scanner.const import SUBSCRIPTABLE
        from .scanner.exceptions import ScanningError
        from .scanner.exceptions import UnresolvedCase
        
        def _get_blocks(lines):
            for match in get_all_blocks(*lines):
                text = match.fulltext.strip()
                if not text:
                    continue
                if not text.startswith('lk.log'):
                    # FIXME (Warning): in lk-logger v4.0.*, the varname-
                    #  detection feature is only enabled when content starts
                    #  with 'lk.log'.
                    continue
                elif re.match(r'^lk\.log\w*\(\)', text):
                    continue
                try:
                    nonlocal node
                    varnames = _analyse_block(text)
                    lineno = match.cursor.lineno + 1
                    node[lineno] = tuple(varnames)
                except ScanningError:
                    nonlocal filename
                    raise ScanningError(
                        match.cursor.lineno + 1, text,
                        0, '<unknown>',
                        f'<filename: {filename}>'
                    )
        
        def _analyse_block(text: str):
            varnames = []
            try:
                for element, type_ in get_variables(text):
                    if type_ == VARIABLE_NAME:
                        varnames.append(element)
                    elif type_ == SUBSCRIPTABLE:
                        varnames.append(_analyse_subscriptables(
                            element, shorten_sub_substrings=True
                        ))
                    else:
                        varnames.append('')
            except UnresolvedCase:
                varnames.clear()
            finally:
                return varnames
                
        def _analyse_subscriptables(
                text: str, shorten_sub_substrings=True, threshold=20
        ):
            """
            shorten_sub_strings:
                example:
                    case 1: when captured a "varname" like "xxx('hello world')":
                        if shorten_sub_strings is True:
                            varname updated: "xxx('...')"
                        if shorten_sub_strings is False:
                            varname updated: "xxx('hello world')"
                    case 2: when captured a "varname" like "xxx('''hello world\n
                            hello world\nhello world\nhello world\nhello...''')"
                            (which is a very long sentense):
                        if shorten_sub_strings is True:
                            varname updated: "xxx('...')"
                        if shorten_sub_strings is False:
                            varname updated: "xxx('''hello world\nhello world\n
                                hello world\nhello world\nhello world\n...)"
                note: the character threshold length to trigger shortening a
                    string is adjustable, the default threshold is 10 chars and
                    must with no line break in it.
            """
            if not shorten_sub_substrings:
                return text
            
            backslash_mask = []
            for i in re.findall(r'\\.', text):
                backslash_mask.append(i)
            if backslash_mask:
                text = re.sub(r'\\.', '__BACKSLASK_MASK__', text)
            
            quotes_pattern_1 = re.compile(r'\'\'\'[\w\W]*\'\'\'|"""[\w\W]*"""')
            quotes_pattern_2 = re.compile(r'\'[^\']*\'|"[^"]*"')
            
            for i in set(quotes_pattern_1.findall(text)):
                if '\n' in i or len(i) > threshold:
                    text = text.replace(i, '"""..."""')
            
            for i in set(quotes_pattern_2.findall(text)):
                if '\n' in i or len(i) > threshold:
                    text = text.replace(i, '"..."')

            # restore backslashes
            if backslash_mask:
                for i in backslash_mask:
                    text = text.replace('__BACKSLASK_MASK__', i, 1)
                del backslash_mask

            text = re.sub(r'\s+', ' ', text)
            #   note: this ^measure takes a side effect that it may replace
            #   sequential whitespaces to one whitespace inside quote strings.

            return text

        node = self._sourcemap.setdefault(filename, {})
        
        with open(filename, 'r', encoding='utf-8') as f:
            node = self._sourcemap.setdefault(filename, {})
            _get_blocks((x.rstrip('\n') for x in f.readlines()))


frame_finder = FrameFinder()
getframe = frame_finder.getframe
sourcemap = SourceMap()

__all__ = ['frame_finder', 'getframe', 'sourcemap']



from ..logger import BaseLogger
from ..sourcemap import getframe


class PycharmLogger(BaseLogger):
    
    @getframe
    def log(self, *data, h='self'):
        self.output(self.format(
            data, advanced=False
        ))
    
    @getframe
    def loga(self, *data, h='self'):
        self.output(self.format(
            data, advanced=True
        ))
    
    @getframe
    def logd(self, *data, symbol='-', length=80, h='self'):
        self.output(self.format(
            data, advanced=True, divider_line=symbol * length
        ))
    
    @getframe
    def logp(self, *data, title='', h='self'):
        from pprint import pformat
        from textwrap import indent
        
        temp = []
        for i, (v, d) in enumerate(
                zip(self._get_varnames() or ['_'] * len(data), data)
        ):
            temp.append(f'[{i}] {v or "_"}:')
            temp.append(indent(pformat(d), '    '))
        
        self.output(self.format(
            temp, advanced=False, sep='\n', title=title,
            start_from_newline=True
        ))
    
    @getframe
    def logt(self, tag, *data, h='self'):
        self.output(self.format(
            data, advanced=True, tag=tag
        ))
    
    @getframe
    def logx(self, *data, h='self'):
        self.output(self.format(
            data, advanced=False, count=self._fmt_update_count()
        ))
    
    @getframe
    def logax(self, *data, h='self'):
        self.output(self.format(
            data, advanced=True, count=self._fmt_update_count()
        ))
    
    @getframe
    def logdx(self, *data, symbol='-', length=80, h='self'):
        self.output(self.format(
            data, advanced=True, divider_line=symbol * length,
            count=self._fmt_update_count()
        ))
    
    @getframe
    def logtx(self, tag, *data, h='self'):
        self.output(self.format(
            data, advanced=True, tag=tag, count=self._fmt_update_count()
        ))
    
    @getframe
    def logdtx(self, tag, *data, symbol='-', length=80, h='self'):
        self.output(self.format(
            data, advanced=True, divider_line=symbol * length, tag=tag,
            count=self._fmt_update_count()
        ))
    
    def output(self, msg: str, **kwargs):
        print(msg)


lk = PycharmLogger()



"""
Abbreviations:
    bak     backup
    fmt     format, formatify
    tmp     template
"""
import re


def fmt_count(count, total_count,
              template='[{i}/{j}]', backup_template='[{i}]'):
    """
    Args:
        count
        total_count
        template:
            optional keys:
                key     example     note
                i       12
                I       012      <- j=100
                        001,112  <- j=0
                I1, I2, I3, ...
                        0012     <- I4
                j       100
                p       1.00%    <- i=1, j=100
                p.0, p.1, p.2, ...
                        1.0000%  <- p.4
                P       01.00%   <- i=1, j=100
                P.0, P.1, P.2, ...
                        01.000%  <- P.3
        backup_template:
            if total count is zero, will use this.

    References:
        https://www.runoob.com/python/att-string-format.html
    """
    
    data = {}
    i, j = count, total_count
    p = i / j if j else ''
    template = template if j else backup_template
    
    for x in _parse_template(template):
        if x == 'i':
            data[x] = i
        elif x == 'I':
            if j:
                data[x] = '{{:0>{}}}'.format(len(str(j))).format(i)
            else:
                dynamic_width = 3
                i_width = len(str(i))
                while dynamic_width < i_width:
                    dynamic_width += 3
                data[x] = '0' * (dynamic_width - i_width) + ':,'.format(i)
        elif x.startswith('I'):
            data[x] = '{{:0>{}}}'.format(int(x[1:])).format(i)
        elif x == 'j' and j:
            data[x] = j
        elif x == 'p' and p:
            data[x] = '{:.2%}'.format(p)
        elif x.startswith('p.') and p:
            if x == 'p.0':
                data[x] = '{}%'.format(round(p))
            else:
                data[x] = '{{:.{}%}}'.format(int(x[2:])).format(p)
        elif x == 'P' and p:
            if p < 10:
                data[x] = '0' + '{:.2%}'.format(p)
            else:
                data[x] = '{:.2%}'.format(p)
        elif x.startswith('P.') and p:
            if p < 10:
                data[x] = '0' + '{{:.{}%}}'.format(int(x[2:])).format(p)
            else:
                data[x] = '{{:.{}%}}'.format(int(x[2:])).format(p)
        else:
            data[x] = ''
            # raise ValueError(x, template)
    
    return template.format(**data)


# def fmt_tag(template='[{t}]'):
#     pass


def ratchet_width():
    pass


def _parse_template(template):
    slot = re.compile(r'(?<={)[^}]+')
    return set(slot.findall(template))



from inspect import currentframe
from textwrap import indent

from .plugins import Counter
from .sourcemap import getframe
from .sourcemap import sourcemap


class BaseLogger(Counter):
    
    def __init__(self, **kwargs):
        super().__init__(auto_reset_count=kwargs.get('auto_reset_count', True))
        
        self.config = kwargs
        
        # TODO: assign the most frequently used configs to directly acessable
        #   attributes.
        self._template = self.config.get(
            'template', '{filename}:{lineno}\t>>\t{func}\t>>\t{msg}')
        self._var_seg = self.config.get('var_seg', ';\t')
        self._visualize_linebreaks = self.config.get(  # DEL
            'visualize_linebreaks', False)
        
        self.mode = 'full_mode'
        self.__mode = {
            'disabled' : (
                lambda *_, **__: None,
                lambda *_, **__: None,
            ),
            'lite_mode': (
                lambda data, **__: ';\t'.join(map(str, data)),
                lambda data, **__: print(data)
            ),
            'full_mode': (
                self.format,
                self.output,
            ),
        }
    
    # -------------------------------------------------------------------------
    # master control
    
    def switch_mode(self, mode: str, quiet=False, _h=1):
        # mode: str['lite_mode', 'full_mode', 'disabled']
        if self.mode == mode:
            return
        else:
            self.mode = mode
            
            if not quiet:
                # reference: `.sourcemap.FrameFinder.getframe._wrap` and
                # `self.position`.
                frame = currentframe()
                for _ in range(_h): frame = frame.f_back
                filename, lineno = frame.f_code.co_filename, frame.f_lineno
                print(f'{filename}:{lineno}\t>>\t'
                      f'[lk_logger] mode switched: {mode}')
            
        a, b = self.__mode[self.mode]
        # tip: here we use `setattr(...)` not `self.format = ...` to avoid
        # PEP-8 (weak) warnings and fix code navigation problem (and some
        # intelli-sense problems) when developing in pycharm.
        setattr(self, 'format', a)
        setattr(self, 'output', b)
    
    def disable(self):
        if self.mode == 'disabled':
            return
        else:
            self.mode = 'disabled'
        a, b = self.__mode[self.mode]
        setattr(self, 'format', a)
        setattr(self, 'output', b)
    
    def enable_lite_mode(self, quiet=False):
        self.switch_mode('lite_mode', quiet=quiet, _h=2)
    
    def disable_lite_mode(self, quiet=False):
        if self.mode == 'disabled':
            raise Exception(
                '[lk_logger.logger.BaseLogger]',
                'lk logger is disabled, you should call `lk.enable()` first to '
                'reactivate master control.'
            )
        self.switch_mode('full_mode', quiet=quiet, _h=2)
    
    # -------------------------------------------------------------------------
    
    def format(self, data, **kwargs):
        """
        
        Args:
            data:
            **kwargs:
        
        Keyword Args:
            advanced
            count
            divider_line
            indent: int[4]
            start_from_newline: bool[False].
            tag:
            title: effect when `start_from_newline` is True.
        """
        info = sourcemap.get_frame_info(
            advanced=kwargs.get('advanced', False)
        )
        
        # message head
        msg_head = ' '.join(filter(None, (
            kwargs.get('tag'),
            kwargs.get('count'),
            kwargs.get('divider_line'),
        ))).strip()
        
        # message body
        if info.varnames:
            if kwargs.get('tag'):
                varnames = info.varnames[1:]
            else:
                varnames = info.varnames
            assert len(varnames) == len(data), (info, data)
            temp = []
            for k, v in zip(varnames, data):
                temp.append(f'{k} = {v}' if k else str(v))
            msg_body = kwargs.get('sep', ';\t').join(temp)
        else:
            msg_body = kwargs.get('sep', ';\t').join(map(str, data))
        if self._visualize_linebreaks:
            msg_body = msg_body.replace('\n', '\\n')
        if kwargs.get('start_from_newline', False):
            title = kwargs.get('title', '')
            msg_body = title + indent(
                '\n' + msg_body, ' ' * kwargs.get('indent', 4)
            )
        
        out = self._template.format(
            # source=f'{info.filename}:{info.lineno}',  # TODO: fixed width
            filename=info.filename,
            lineno=info.lineno,
            func=info.name,
            msg=f'{msg_head} {msg_body}'.strip(' ')
        )
        return out
    
    @staticmethod
    def _get_varnames():
        info = sourcemap.get_frame_info(advanced=True)
        return info.varnames
    
    # -------------------------------------------------------------------------
    # Typical implementations see: `lk_logger.terminals.pycharm_console`.
    
    @getframe
    def log(self, *data, h='self'):
        raise NotImplementedError
    
    @getframe
    def loga(self, *data, h='self'):
        raise NotImplementedError
    
    @getframe
    def logd(self, *data, symbol='-', length=80, h='self'):
        raise NotImplementedError
    
    @getframe
    def logp(self, *data, h='self'):
        raise NotImplementedError
    
    @getframe
    def logt(self, tag, *data, h='self'):
        raise NotImplementedError
    
    @getframe
    def logx(self, *data, h='self'):
        raise NotImplementedError
    
    @getframe
    def logtx(self, tag, *data, h='self'):
        raise NotImplementedError
    
    @getframe
    def logax(self, *data, h='self'):
        raise NotImplementedError
    
    @getframe
    def logdx(self, *data, symbol='-', length=80, h='self'):
        raise NotImplementedError
    
    @getframe
    def logtx(self, *data, h='self'):
        raise NotImplementedError
    
    @getframe
    def logdtx(self, tag, *data, symbol='-', length=80, h='self'):
        raise NotImplementedError
    
    def output(self, msg: str, **kwargs):
        raise NotImplementedError
    
    # -------------------------------------------------------------------------
    
    @property
    @getframe
    def position(self):
        """
        
        Returns:
            (filename, lineno)
                - filename is an absolute path.
                - the lineno starts from 1.
        """
        from .sourcemap import frame_finder
        frame = frame_finder.frame
        return frame.f_code.co_filename, frame.f_lineno
    
    @property
    def version(self):
        from . import __version__
        return __version__



from time import time


class Timer:
    
    def __init__(self):
        self._start_time = time()
        self._end_time = None
    
    def start_timer(self):
        self._start_time = time()
    
    def end_timer(self):
        self._end_time = time()
    
    @property
    def _elapsed_time(self):
        return self._end_time and self._end_time - self._start_time
